use crate::*;
use rand::prelude::*;
use std::{collections::HashMap, fmt};

// MUST be less than total amount of letters divided by two
const CABLES: usize = 10;

#[derive(Debug, Clone)]
pub struct Plugboard {
    mappings: Vec<Letter>,
}

impl Plugboard {
    pub fn new(mappings: Vec<Letter>) -> Self {
        Plugboard { mappings }
    }

    pub fn identity() -> Self {
        let mappings = vec![
            Letter::A,
            Letter::B,
            Letter::C,
            Letter::D,
            Letter::E,
            Letter::F,
            Letter::G,
            Letter::H,
            Letter::I,
            Letter::J,
            Letter::K,
            Letter::L,
            Letter::M,
            Letter::N,
            Letter::O,
            Letter::P,
            Letter::Q,
            Letter::R,
            Letter::S,
            Letter::T,
            Letter::U,
            Letter::V,
            Letter::W,
            Letter::X,
            Letter::Y,
            Letter::Z,
        ];

        Plugboard::new(mappings)
    }

    pub fn index(&self, index: usize) -> usize {
        *self.mappings.get(index).expect(
            &format!(
                "Switchboard is missing mapping for a letter: {:?}",
                Letter::from(index)
            )
            .to_string(),
        ) as usize
    }

    pub fn letter(&self, letter: Letter) -> Letter {
        Letter::from(self.index(letter as usize))
    }
}

impl fmt::Display for Plugboard {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let mut pairs = HashMap::new();

        for (i, mapping) in self.mappings.iter().enumerate() {
            let left_letter = Letter::from(i);
            let right_letter = *mapping;

            if left_letter != right_letter
                && !pairs.get(&left_letter).is_some()
                && !pairs.get(&right_letter).is_some()
            {
                pairs.insert(left_letter, right_letter);
            }
        }

        let mut s = String::new();
        if pairs.is_empty() {
            s = "Identity".to_string();
        }

        for (left_letter, right_letter) in pairs.iter() {
            s = format!("{:?}{:?} {}", left_letter, right_letter, s);
        }

        f.write_str(&s)
    }
}

impl Random for Plugboard {
    fn random() -> Self {
        let mut mappings = vec![Letter::A; AMOUNT_OF_LETTERS];
        let mut rng = rand::thread_rng();

        let mut input = Letter::all();
        input.shuffle(&mut rng);
        input.truncate(CABLES);

        let mut output: Vec<Letter> = Letter::all()
            .iter()
            .filter(|l| !input.contains(l))
            .cloned()
            .collect();
        let remaining = output.split_off(CABLES);

        for (input, output) in input.iter().zip(output.iter()) {
            mappings.insert(*input as usize, *output);
            mappings.insert(*output as usize, *input);
        }

        for letter in remaining {
            mappings.insert(letter as usize, letter);
        }

        Plugboard::new(mappings)
    }
}
